"""
==============
Flow Executors
==============

.. autoclass:: resolwe.flow.executors.BaseFlowExecutor
    :members:

"""
from __future__ import absolute_import, division, print_function, unicode_literals

import json
import logging
import os
import shutil
import six
import traceback
import uuid

from django.apps import apps
from django.db import transaction
from django.conf import settings

from resolwe.flow.models import Data, dict_dot, iterate_fields, Process
from resolwe.flow.utils.purge import data_purge
from resolwe.utils import BraceMessage as __

if settings.USE_TZ:
    from django.utils.timezone import now
else:
    import datetime
    now = datetime.datetime.now  # pylint: disable=invalid-name


logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
CWD = os.getcwd()

exported_files_mapper = {}


def iterjson(text):
    """Decode JSON stream."""
    decoder = json.JSONDecoder()
    while len(text) > 0:
        obj, ndx = decoder.raw_decode(text)

        if not isinstance(obj, dict):
            raise ValueError()

        text = text[ndx:].lstrip('\r\n')
        yield obj


def hydrate_spawned_files(filename, data_id):
    if filename not in exported_files_mapper:
        raise KeyError('All files referenced in spawned processes must be exported using'
                       '`re-export` command.')

    export_fn = exported_files_mapper[filename]
    return {'file_temp': export_fn, 'file': filename}


class BaseFlowExecutor(object):

    """Represents a workflow executor

    .. WARNING::
        Ensure this class can be serialized since some engines will
        use it this way (e.g. Celery).
    """

    def get_tools(self):
        tools_paths = []
        for app_config in apps.get_app_configs():
            proc_path = os.path.join(app_config.path, 'tools')
            if os.path.isdir(proc_path):
                tools_paths.append(proc_path)

        custom_tools_paths = getattr(settings, 'RESOLWE_CUSTOM_TOOLS_PATHS', [])
        if not isinstance(custom_tools_paths, list):
            raise KeyError("`RESOLWE_CUSTOM_TOOLS_PATHS` setting must be a list.")
        tools_paths.extend(custom_tools_paths)

        return tools_paths

    def start(self):
        pass

    def run_script(self, script):
        raise NotImplementedError('`run_script` function must be defined')

    def end(self):
        pass

    def get_stdout(self):
        return self.stdout  # pylint: disable=no-member

    def update_data_status(self, **kwargs):
        data = Data.objects.get(pk=self.data_id)
        for key, value in kwargs.items():
            setattr(data, key, value)

        # Ensure that we only update the fields that were changed.
        data.save(update_fields=kwargs.keys())

    def run(self, data_id, script, verbosity=1):
        """Execute the script and save results."""
        if verbosity >= 1:
            print('RUN: {} {}'.format(data_id, script))

        self.data_id = data_id

        data_dir = settings.FLOW_EXECUTOR['DATA_DIR']
        dir_mode = getattr(settings, 'FLOW_EXECUTOR', {}).get('DATA_DIR_MODE', 0o755)

        output_path = os.path.join(data_dir, str(data_id))

        os.mkdir(output_path)
        # os.mkdir is not guaranteed to set the given mode
        os.chmod(output_path, dir_mode)
        os.chdir(output_path)

        log_file = open('stdout.txt', 'w+')
        json_file = open('jsonout.txt', 'w+')

        proc_pid = self.start()

        self.update_data_status(
            status=Data.STATUS_PROCESSING,
            started=now(),
            process_pid=proc_pid
        )

        # Run processor and handle intermediate results
        self.run_script(script)
        spawn_processors = []
        output = {}
        process_error, process_warning, process_info = [], [], []
        process_progress, process_rc = 0, 0

        # read processor output
        try:
            stdout = self.get_stdout()
            while True:
                line = stdout.readline()
                if not line:
                    break

                try:
                    if line.strip().startswith('run'):
                        # Save processor and spawn if no errors
                        log_file.write(line)
                        log_file.flush()

                        for obj in iterjson(line[3:].strip()):
                            spawn_processors.append(obj)
                    elif line.strip().startswith('export'):
                        file_name = line[6:].strip()

                        export_folder = settings.FLOW_EXECUTOR['UPLOAD_DIR']
                        unique_name = 'export_{}'.format(uuid.uuid4().hex)
                        export_path = os.path.join(export_folder, unique_name)

                        exported_files_mapper[file_name] = unique_name

                        shutil.move(file_name, export_path)
                    else:
                        # If JSON, save to MongoDB
                        updates = {}
                        for obj in iterjson(line):
                            for key, val in six.iteritems(obj):
                                if key.startswith('proc.'):
                                    if key == 'proc.error':
                                        process_error.append(val)
                                        if not process_rc:
                                            process_rc = 1
                                            updates['process_rc'] = process_rc
                                        updates['process_error'] = process_error
                                        updates['status'] = Data.STATUS_ERROR
                                    elif key == 'proc.warning':
                                        process_warning.append(val)
                                        updates['process_warning'] = process_warning
                                    elif key == 'proc.info':
                                        process_info.append(val)
                                        updates['process_info'] = process_info
                                    elif key == 'proc.rc':
                                        process_rc = int(val)
                                        updates['process_rc'] = process_rc
                                        if process_rc != 0:
                                            updates['status'] = Data.STATUS_ERROR
                                    elif key == 'proc.progress':
                                        process_progress = int(float(val) * 100)
                                        updates['process_progress'] = process_progress
                                else:
                                    dict_dot(output, key, val)
                                    updates['output'] = output

                        if updates:
                            updates['modified'] = now()
                            self.update_data_status(**updates)

                        if process_rc > 0:
                            log_file.close()
                            json_file.close()
                            os.chdir(CWD)
                            return

                        # Debug output
                        # Not referenced in Data object
                        json_file.write(line)
                        json_file.flush()

                except ValueError as ex:
                    # Ignore if not JSON
                    log_file.write(line)
                    log_file.flush()

        except MemoryError as ex:
            logger.error(__("Out of memory: {}", ex))

        except IOError as ex:
            # TODO: if ex.errno == 28: no more free space
            raise ex
        finally:
            # Store results
            log_file.close()
            json_file.close()
            os.chdir(CWD)

        return_code = self.end()

        if process_rc < return_code:
            process_rc = return_code

        if spawn_processors and process_rc == 0:
            parent_data = Data.objects.get(pk=self.data_id)

            # Spawn processors
            for d in spawn_processors:
                d['contributor'] = parent_data.contributor
                d['process'] = Process.objects.get(slug=d['process'])

                for field_schema, fields in iterate_fields(d.get('input', {}), d['process'].input_schema):
                    type_ = field_schema['type']
                    name = field_schema['name']
                    value = fields[name]

                    if type_ == 'basic:file:':
                        fields[name] = hydrate_spawned_files(value, data_id)
                    elif type_ == 'list:basic:file:':
                        fields[name] = [hydrate_spawned_files(fn, data_id) for fn in value]

                with transaction.atomic():
                    d = Data.objects.create(**d)
                    for collection in parent_data.collection_set.all():
                        collection.data.add(d)

        if process_rc == 0:
            self.update_data_status(
                status=Data.STATUS_DONE,
                process_progress=100,
                finished=now()
            )
        else:
            self.update_data_status(
                status=Data.STATUS_ERROR,
                process_progress=100,
                process_rc=process_rc,
                finished=now()
            )

        try:
            # Cleanup after processor
            if data_id != 'no_data_id':
                data_purge(data_ids=[data_id], delete=True, verbosity=verbosity)
        except:  # pylint: disable=bare-except
            logger.error(__("Purge error:\n\n{}", traceback.format_exc()))

        # if not update_data(data):  # Data was deleted
        #     # Restore original directory
        #     os.chdir(settings.PROJECT_ROOT)
        #     return

        # Restore original directory
        # os.chdir(settings.PROJECT_ROOT)

        # return data_id

    def terminate(self, data_id):
        """Terminate a running script."""
        raise NotImplementedError('subclasses of BaseFlowExecutor may require a terminate() method')



"""
=======
Testing
=======

"""
from __future__ import absolute_import, division, print_function, unicode_literals

import copy
import hashlib
import gzip
import io
import json
import os
import shutil
from six.moves import filterfalse
import zipfile

from django.conf import settings
from django.contrib.auth import get_user_model
from django.core import management
from django.test import TestCase, override_settings
from django.utils.crypto import get_random_string
from django.utils.text import slugify

from resolwe.flow.models import Data, dict_dot, iterate_fields, Collection, DescriptorSchema, Process, Storage
from resolwe.flow.managers import manager


SCHEMAS_FIXTURE_CACHE = None


# override all FLOW_EXECUTOR settings that are specified in FLOW_EXECUTOR['TEST']
flow_executor_settings = copy.copy(getattr(settings, 'FLOW_EXECUTOR', {}))
test_settings_overrides = getattr(settings, 'FLOW_EXECUTOR', {}).get('TEST', {})
flow_executor_settings.update(test_settings_overrides)

# update FLOW_DOCKER_MAPPINGS setting if necessary
flow_docker_mappings = copy.copy(getattr(settings, 'FLOW_DOCKER_MAPPINGS', {}))
for map_ in flow_docker_mappings:
    for map_entry in ['src', 'dest']:
        for setting in ['DATA_DIR', 'UPLOAD_DIR']:
            if settings.FLOW_EXECUTOR[setting] in map_[map_entry]:
                map_[map_entry] = map_[map_entry].replace(
                    settings.FLOW_EXECUTOR[setting], flow_executor_settings[setting], 1)


@override_settings(FLOW_EXECUTOR=flow_executor_settings)
@override_settings(FLOW_DOCKER_MAPPINGS=flow_docker_mappings)
@override_settings(CELERY_ALWAYS_EAGER=True)
class ProcessTestCase(TestCase):

    """Base class for writing process tests.

    This class is subclass of Django's :class:`~django.test.TestCase`
    with some specific functions used for testing processes.

    To write a process test use standard Django's syntax for writing
    tests and follow the next steps:

    #. Put input files (if any) in ``tests/files`` directory of a
       Django application.
    #. Run the process using :meth:`run_process`.
    #. Check if the process has the expected status using
       :meth:`assertStatus`.
    #. Check process's output using :meth:`assertFields`,
       :meth:`assertFile`, :meth:`assertFileExists`,
       :meth:`assertFiles` and :meth:`assertJSON`.

    .. note::
        When creating a test case for a custom Django application,
        subclass this class and over-ride the ``self.files_path`` with:

        .. code-block:: python

            self.files_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'files')

    .. DANGER::
        If output files don't exist in ``tests/files`` directory of a
        Django application, they are created automatically.
        But you have to check that they are correct before using them
        for further runs.

    """

    def _register_schemas(self, **kwargs):
        """Register process and descriptor schemas.

        Process and DescriptorSchema cached to SCHEMAS_FIXTURE_CACHE
        global variable based on ``path`` key in ``kwargs``.

        """
        Process.objects.all().delete()

        cache_key = json.dumps(kwargs)

        global SCHEMAS_FIXTURE_CACHE  # pylint: disable=global-statement
        if not SCHEMAS_FIXTURE_CACHE:
            SCHEMAS_FIXTURE_CACHE = {}

        if cache_key in SCHEMAS_FIXTURE_CACHE:
            Process.objects.bulk_create(SCHEMAS_FIXTURE_CACHE[cache_key]['processes'])
            DescriptorSchema.objects.bulk_create(SCHEMAS_FIXTURE_CACHE[cache_key]['descriptor_schemas'])
        else:
            user_model = get_user_model()

            if not user_model.objects.filter(is_superuser=True).exists():
                user_model.objects.create_superuser(
                    username="admin", email='admin@example.com', password="admin_pass")

            management.call_command('register', force=True, testing=True, verbosity='0', **kwargs)

            if cache_key not in SCHEMAS_FIXTURE_CACHE:
                SCHEMAS_FIXTURE_CACHE[cache_key] = {}

            # list forces db query execution
            SCHEMAS_FIXTURE_CACHE[cache_key]['processes'] = list(Process.objects.all())
            SCHEMAS_FIXTURE_CACHE[cache_key]['descriptor_schemas'] = list(DescriptorSchema.objects.all())

    def setUp(self):
        super(ProcessTestCase, self).setUp()
        self.admin = get_user_model().objects.create_superuser(
            username="admin", email='admin@example.com', password="admin_pass")
        self._register_schemas()

        self.collection = Collection.objects.create(contributor=self.admin, name="Test collection")
        self.files_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'files')
        self.upload_dir = settings.FLOW_EXECUTOR['UPLOAD_DIR']
        self._keep_all = False
        self._keep_failed = False
        self._upload_files = []

        # create upload dir if it doesn't exist
        if not os.path.isdir(self.upload_dir):
            os.mkdir(self.upload_dir)

    def tearDown(self):
        super(ProcessTestCase, self).tearDown()

        # Delete Data objects and their files unless keep_failed
        for d in Data.objects.all():
            if self._keep_all or (self._keep_failed and d.status == "error"):
                print("KEEPING DATA: {}".format(d.pk))
            else:
                data_dir = os.path.join(settings.FLOW_EXECUTOR['DATA_DIR'], str(d.pk))
                export_dir = os.path.join(settings.FLOW_EXECUTOR['UPLOAD_DIR'], str(d.pk))
                d.delete()
                shutil.rmtree(data_dir, ignore_errors=True)
                shutil.rmtree(export_dir, ignore_errors=True)

        # remove uploaded files
        if not self._keep_all and not self._keep_failed:
            for fn in self._upload_files:
                shutil.rmtree(fn, ignore_errors=True)

    def keep_all(self):
        """Do not delete output files after test for all data."""
        self._keep_all = True

    def keep_failed(self):
        """Do not delete output files after test for failed data."""
        self._keep_failed = True

    def run_processor(self, *args, **kwargs):
        return self.run_process(*args, **kwargs)
        # TODO: warning

    def run_process(self, process_slug, input_={}, assert_status=Data.STATUS_DONE,
                    descriptor=None, descriptor_schema=None, run_manager=True,
                    verbosity=0):
        """Run the specified process with the given inputs.

        If input is a file, file path should be given relative to the
        ``tests/files`` directory of a Django application.
        If ``assert_status`` is given, check if
        :class:`~resolwe.flow.models.Data` object's status matches
        it after the process has finished.

        :param str process_slug: slug of the
            :class:`~resolwe.flow.models.Process` to run

        :param dict ``input_``: :class:`~resolwe.flow.models.Process`'s
            input parameters

            .. note::

                You don't have to specify parameters with defined
                default values.

        :param str ``assert_status``: desired status of the
            :class:`~resolwe.flow.models.Data` object

        :param dict descriptor: descriptor to set on the
            :class:`~resolwe.flow.models.Data` object

        :param dict descriptor_schema: descriptor schema to set on the
            :class:`~resolwe.flow.models.Data` object

        :return: object created by
            :class:`~resolwe.flow.models.Process`
        :rtype: ~resolwe.flow.models.Data

        """

        # backward compatibility
        process_slug = slugify(process_slug.replace(':', '-'))

        p = Process.objects.get(slug=process_slug)

        def mock_upload(file_path):
            old_path = os.path.join(self.files_path, file_path)
            if not os.path.isfile(old_path):
                raise RuntimeError('Missing file: {}'.format(old_path))

            new_path = os.path.join(self.upload_dir, file_path)
            # create directories needed by new_path
            new_path_dir = os.path.dirname(new_path)
            if not os.path.exists(new_path_dir):
                os.makedirs(new_path_dir)
            shutil.copy2(old_path, new_path)
            self._upload_files.append(new_path)
            return {
                'file': file_path,
                'file_temp': file_path,
            }

        for field_schema, fields in iterate_fields(input_, p.input_schema):
            # copy referenced files to upload dir
            if field_schema['type'] == "basic:file:":
                fields[field_schema['name']] = mock_upload(fields[field_schema['name']])
            elif field_schema['type'] == "list:basic:file:":
                file_list = [mock_upload(file_path) for file_path in fields[field_schema['name']]]
                fields[field_schema['name']] = file_list

            # convert primary keys to strings
            if field_schema['type'].startswith('data:'):
                fields[field_schema['name']] = fields[field_schema['name']]
            if field_schema['type'].startswith('list:data:'):
                fields[field_schema['name']] = [obj for obj in fields[field_schema['name']]]

        d = Data.objects.create(
            input=input_,
            contributor=self.admin,
            process=p,
            slug=get_random_string(length=6),
            descriptor_schema=descriptor_schema,
            descriptor=descriptor or {})
        self.collection.data.add(d)

        if run_manager:
            manager.communicate(run_sync=True, verbosity=verbosity)

        # Fetch latest Data object from database
        d = Data.objects.get(pk=d.pk)
        if not run_manager and assert_status == Data.STATUS_DONE:
            assert_status = Data.STATUS_RESOLVING

        if assert_status:
            self.assertStatus(d, assert_status)

        return d

    def assertStatus(self, obj, status):  # pylint: disable=invalid-name
        """Check if object's status is equal to the given status.

        :param obj: object for which to check the status
        :type obj: ~resolwe.flow.models.Data
        :param str status: desired value of object's
            :attr:`~resolwe.flow.models.Data.status` attribute

        """
        self.assertEqual(obj.status, status,
                         msg="Data status is '{}', not '{}'".format(obj.status, status) +
                         self._debug_info(obj))

    def assertFields(self, obj, path, value):  # pylint: disable=invalid-name
        """Compare object's field to the given value.

        :param obj: object with the field to compare
        :type obj: ~resolwe.flow.models.Data

        :param str path: path to
            :class:`~resolwe.flow.models.Data` object's field

        :param str value: desired value of
            :class:`~resolwe.flow.models.Data` object's field

        """
        field = dict_dot(obj.output, path)
        self.assertEqual(field, value,
                         msg="Field 'output.{}' mismatch: {} != {}".format(path, field, value) +
                         self._debug_info(obj))

    def _assert_file(self, obj, fn_tested, fn_correct, compression=None, filter=lambda _: False):
        open_kwargs = {}
        if compression is None:
            open_fn = open
            # by default, open() will open files as text and return str
            # objects, but we need bytes objects
            open_kwargs['mode'] = 'rb'
        elif compression == 'gzip':
            open_fn = gzip.open
        elif compression == 'zip':
            open_fn = zipfile.ZipFile.open
        else:
            raise ValueError("Unsupported compression format.")

        output = os.path.join(settings.FLOW_EXECUTOR['DATA_DIR'], str(obj.pk), fn_tested)
        with open_fn(output, **open_kwargs) as output_file:
            output_contents = b"".join([line for line in filterfalse(filter, output_file)])
        output_hash = hashlib.sha256(output_contents).hexdigest()

        correct_path = os.path.join(self.files_path, fn_correct)

        if not os.path.isfile(correct_path):
            shutil.copyfile(output, correct_path)
            self.fail(msg="Output file {} missing so it was created.".format(fn_correct))

        with open_fn(correct_path, **open_kwargs) as correct_file:
            correct_contents = b"".join([line for line in filterfalse(filter, correct_file)])
        correct_hash = hashlib.sha256(correct_contents).hexdigest()
        self.assertEqual(correct_hash, output_hash,
                         msg="File contents hash mismatch: {} != {}".format(
                             correct_hash, output_hash) + self._debug_info(obj))

    def assertFile(self, obj, field_path, fn, **kwargs):  # pylint: disable=invalid-name
        """Compare a process's output file to the given correct file.

        :param obj: object that includes the file to compare
        :type obj: ~resolwe.flow.models.Data

        :param str field_path: path to
            :class:`~resolwe.flow.models.Data` object's field with the
            file name

        :param str fn: file name (and relative path) of the correct
            file to compare against. Path should be relative to the
            ``tests/files`` directory of a Django application.

        :param str compression: if not ``None``, files will be
            uncompressed with the appropriate compression library
            before comparison.
            Currently supported compression formats are *gzip* and
            *zip*.

        :param filter: function for filtering the contents of output
            files. It is used in :func:`itertools.filterfalse` function
            and takes one parameter, a line of the output file. If it
            returns ``True``, the line is excluded from comparison of
            the two files.
        :type filter: ~types.FunctionType

        """
        field = dict_dot(obj.output, field_path)
        self._assert_file(obj, field['file'], fn, **kwargs)

    def assertFiles(self, obj, field_path, fn_list, **kwargs):  # pylint: disable=invalid-name
        """Compare a list of process's output files to the given
        correct files.

        :param obj: object which includes the files to compare
        :type obj: ~resolwe.flow.models.Data

        :param str field_path: path to
            :class:`~resolwe.flow.models.Data` object's field with the
            list of file names

        :param list fn_list: list of file names (and relative paths) of
            files to compare against. Paths should be relative to the
            ``tests/files`` directory of a Django application.

        :param str compression: if not ``None``, files will be
            uncompressed with the appropriate compression library
            before comparison.
            Currently supported compression formats are *gzip* and
            *zip*.

        :param filter: Function for filtering the contents of output
            files. It is used in :obj:`itertools.filterfalse` function
            and takes one parameter, a line of the output file. If it
            returns ``True``, the line is excluded from comparison of
            the two files.
        :type filter: ~types.FunctionType

        """
        field = dict_dot(obj.output, field_path)

        if len(field) != len(fn_list):
            self.fail(msg="Lengths of list:basic:file field and files list are not equal.")

        for fn_tested, fn_correct in zip(field, fn_list):
            self._assert_file(obj, fn_tested['file'], fn_correct, **kwargs)

    def assertFileExists(self, obj, field_path):  # pylint: disable=invalid-name
        """Ensure a file in the given object's field exists.

        :param obj: object that includes the file for which to check if
            it exists
        :type obj: ~resolwe.flow.models.Data

        :param str field_path: path to
            :class:`~resolwe.flow.models.Data` object's field with the
            file name/path
        """
        field = dict_dot(obj.output, field_path)
        output = os.path.join(settings.FLOW_EXECUTOR['DATA_DIR'], str(obj.pk), field['file'])

        if not os.path.isfile(output):
            self.fail(msg="File {} does not exist.".format(field_path))

    def assertJSON(self, obj, storage, field_path, file_name):  # pylint: disable=invalid-name
        """Compare JSON in Storage object to the given correct JSON.

        :param obj: object to which the
            :class:`~resolwe.flow.models.Storage` object belongs
        :type obj: ~resolwe.flow.models.Data

        :param storage: object or id which contains JSON to compare
        :type storage: :class:`~resolwe.flow.models.Storage` or
            :class:`str`

        :param str field_path: path to JSON subset in the
            :class:`~resolwe.flow.models.Storage`'s object to compare
            against. If it is empty, the entire object will be
            compared.

        :param str file_name: file name (and relative path) of the file
            with the correct JSON to compare against. Path should be
            relative to the ``tests/files`` directory of a Django
            application.

            .. note::

                The given JSON file should be compresed with *gzip* and
                have the ``.gz`` extension.

        """
        self.assertEqual(os.path.splitext(file_name)[1], '.gz', msg='File extension must be .gz')

        if not isinstance(storage, Storage):
            storage = Storage.objects.get(pk=storage)

        storage_obj = dict_dot(storage.json, field_path)

        file_path = os.path.join(self.files_path, file_name)
        if not os.path.isfile(file_path):
            with gzip.open(file_path, mode='wt') as f:
                json.dump(storage_obj, f)

            self.fail(msg="Output file {} missing so it was created.".format(file_name))

        with gzip.open(file_path, mode='rt') as f:
            file_obj = json.load(f)

        self.assertEqual(storage_obj, file_obj,
                         msg="Storage {} field '{}' does not match file {}".format(
                             storage.id, field_path, file_name) + self._debug_info(obj))

    def _debug_info(self, data):
        """Return data's debugging information."""
        msg_header = "Debugging information for data object {}".format(data.pk)
        msg = "\n\n" + len(msg_header) * "=" + "\n" + msg_header + "\n" + len(msg_header) * "=" + "\n"
        path = os.path.join(settings.FLOW_EXECUTOR['DATA_DIR'], str(data.pk), "stdout.txt")
        if os.path.isfile(path):
            msg += "\nstdout.txt:\n" + 11 * "-" + "\n"
            with io.open(path, mode='rt') as fn:
                msg += fn.read()

        if data.process_error:
            msg += "\nProcess' errors:\n" + 16 * "-" + "\n"
            msg += "\n".join(data.process_error)

        if data.process_warning:
            msg += "\nProcess' warnings:\n" + 18 * "-" + "\n"
            msg += "\n".join(data.process_warning)

        return msg



"""
=========================
Delete Unreferenced Files
=========================

"""
from django.core.management.base import BaseCommand

from resolwe.flow.utils.purge import data_purge


class Command(BaseCommand):

    """Purge files with no reference in
    :class:`~resolwe.flow.models.Data` objects.

    """

    help = "Purge files with no reference in Data objects."

    def add_arguments(self, parser):
        parser.add_argument('-d', '--data', type=str, nargs='+', help="list of data ids")
        parser.add_argument('-f', '--force', action='store_true', help="delete unreferenced files")

    def handle(self, *args, **options):
        """Call :func:`~resolwe.flow.utils.purge.data_purge` to purge
        files.

        """
        data_purge(options['data'], options['force'])



"""Local workflow executor"""
from __future__ import absolute_import, division, print_function, unicode_literals

import os
import random
import shlex
import subprocess

from django.conf import settings

from .local import FlowExecutor as LocalFlowExecutor


class FlowExecutor(LocalFlowExecutor):

    def __init__(self, *args, **kwargs):
        super(FlowExecutor, self).__init__(*args, **kwargs)
        self.command = getattr(settings, 'FLOW_EXECUTOR', {}).get('COMMAND', 'docker')

    def start(self):
        # arguments passed to the Docker command
        command_args = {
            'command': self.command,
            'container_image': settings.FLOW_EXECUTOR['CONTAINER_IMAGE'],
        }

        if self.data_id != 'no_data_id':
            data_id = self.data_id
        else:
            # set random container name for tests
            data_id = 'test_{}'.format(random.randint(1000, 9999))
        command_args['container_name'] = '--name=resolwe_{}'.format(data_id)

        # render Docker mappings in FLOW_DOCKER_MAPPINGS setting
        mappings_template = getattr(settings, 'FLOW_DOCKER_MAPPINGS', [])
        context = {'data_id': self.data_id}
        mappings = [{key.format(**context): value.format(**context) for key, value in template.items()}
                    for template in mappings_template]

        # create mappings for tools
        # NOTE: To prevent processes tampering with tools, all tools are mounted read-only
        self.mappings_tools = [{'src': tool, 'dest': '/usr/local/bin/resolwe/{}'.format(i), 'mode': 'ro'}
                               for i, tool in enumerate(self.get_tools())]
        mappings += self.mappings_tools
        # create Docker --volume parameters from mappings
        command_args['volumes'] = ' '.join(['--volume="{src}":"{dest}":{mode}'.format(**map_)
                                            for map_ in mappings])

        # set working directory inside the container to the mapped directory of
        # the current Data's directory
        command_args['workdir'] = ''
        for template in mappings_template:
            if '{data_id}' in template['src']:
                command_args['workdir'] = '--workdir={}'.format(template['dest'])

        # create environment variables to pass certain information to the
        # process running in the container
        command_args['envs'] = ' '.join(['--env={name}={value}'.format(**env) for env in [
            {'name': 'HOST_UID', 'value': os.getuid()},
            {'name': 'HOST_GID', 'value': os.getgid()},
        ]])

        # a login Bash shell is needed to source ~/.bash_profile
        command_args['shell'] = '/bin/bash --login'

        self.proc = subprocess.Popen(
            shlex.split(
                '{command} run --rm --interactive {container_name} {volumes} '
                '{envs} {workdir} {container_image} {shell}'.format(**command_args)),
            stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
            universal_newlines=True)

        self.stdout = self.proc.stdout

    def run_script(self, script):
        mappings = getattr(settings, 'FLOW_DOCKER_MAPPINGS', {})
        for map_ in mappings:
            script = script.replace(map_['src'], map_['dest'])
        # create a Bash command to add all the tools to PATH
        tools_paths = ':'.join([map_["dest"] for map_ in self.mappings_tools])
        add_tools_path = 'export PATH=$PATH:{}'.format(tools_paths)
        self.proc.stdin.write(os.linesep.join(['set -x', 'set +B', add_tools_path, script]) + os.linesep)
        self.proc.stdin.close()

    def end(self):
        self.proc.wait()

        return self.proc.returncode

    def terminate(self, data_id):
        subprocess.call(shlex.split('{} rm -f {}'.format(self.command, data_id)))



from __future__ import absolute_import, division, print_function, unicode_literals

from rest_framework.filters import BaseFilterBackend

from resolwe.permissions.shortcuts import get_objects_for_user


class ResolwePermissionsFilter(BaseFilterBackend):

    def filter_queryset(self, request, queryset, view):
        user = request.user

        app_label = queryset.model._meta.app_label
        model_name = queryset.model._meta.model_name

        kwargs = {}
        if model_name == 'storage':
            model_name = 'data'
            kwargs['perms_filter'] = 'data__pk__in'

        permission = '{}.view_{}'.format(app_label, model_name)

        return get_objects_for_user(user, permission, queryset, **kwargs)



# pylint: disable=missing-docstring
from __future__ import absolute_import, division, print_function, unicode_literals

import mock
import os
import shlex
import six
import subprocess
import unittest

from django.conf import settings
from django.test import override_settings

from resolwe.flow.executors.docker import FlowExecutor
from resolwe.flow.executors import BaseFlowExecutor
from resolwe.flow.models import Data
from resolwe.flow.utils.test import ProcessTestCase

try:
    import builtins  # py3
except ImportError:
    import __builtin__ as builtins  # py2


PROCESSES_DIR = os.path.join(os.path.dirname(__file__), 'processes')


def check_docker():
    """Check if Docker is installed and working.

    :return: tuple (indicator of the availability of Docker, reason for
             unavailability)
    :rtype: (bool, str)

    """
    command = getattr(settings, 'FLOW_EXECUTOR', {}).get('COMMAND', 'docker')
    info_command = '{} info'.format(command)
    available, reason = True, ""
    # TODO: use subprocess.DEVNULL after dropping support for Python 2
    with open(os.devnull, 'wb') as DEVNULL:
        try:
            subprocess.check_call(shlex.split(info_command), stdout=DEVNULL, stderr=subprocess.STDOUT)
        except OSError:
            available, reason = False, "Docker command '{}' not found".format(command)
        except subprocess.CalledProcessError:
            available, reason = (False, "Docker command '{}' returned non-zero "
                                        "exit status".format(info_command))
    return available, reason


class DockerExecutorTestCase(unittest.TestCase):

    @unittest.skipUnless(*check_docker())
    @mock.patch('os.mkdir')
    @mock.patch('os.chmod')
    @mock.patch('os.chdir')
    @mock.patch('resolwe.flow.executors.Data.objects.filter')
    @mock.patch('resolwe.flow.executors.Data.objects.get')
    def test_run_in_docker(self, data_get_mock, data_filter_mock, chdir_mock, chmod_mock, mkdir_mock):
        executor_settings = settings.FLOW_EXECUTOR
        executor_settings['CONTAINER_IMAGE'] = 'centos'

        with override_settings(FLOW_EXECUTOR=executor_settings):
            executor = FlowExecutor()

            script = 'if grep -Fq "docker" /proc/1/cgroup; then echo "Running inside Docker"; ' \
                    'else echo "Running locally"; fi'

            count = {'running': 0}

            def assert_output(line):
                if line.strip() == 'Running inside Docker':
                    count['running'] += 1

            write_mock = mock.MagicMock(side_effect=assert_output)
            stdout_mock = mock.MagicMock(write=write_mock)
            open_mock = mock.MagicMock(side_effect=[stdout_mock, mock.MagicMock()])
            with mock.patch.object(builtins, 'open', open_mock):
                executor.run('no_data_id', script, verbosity=0)

            self.assertEqual(count['running'], 1)


class GetToolsTestCase(unittest.TestCase):
    @mock.patch('resolwe.flow.executors.apps')
    @mock.patch('resolwe.flow.executors.os')
    @mock.patch('resolwe.flow.executors.settings')
    def test_get_tools(self, settings_mock, os_mock, apps_mock):
        apps_mock.get_app_configs.return_value = [
            mock.MagicMock(path='/resolwe/test_app1'),
            mock.MagicMock(path='/resolwe/test_app2'),
        ]
        os_mock.path.join = os.path.join
        os_mock.path.isdir.side_effect = [False, True]
        settings_mock.RESOLWE_CUSTOM_TOOLS_PATHS = ['/custom_tools']

        base_executor = BaseFlowExecutor()
        tools_list = base_executor.get_tools()

        self.assertEqual(len(tools_list), 2)
        self.assertIn('/resolwe/test_app2/tools', tools_list)
        self.assertIn('/custom_tools', tools_list)

    @mock.patch('resolwe.flow.executors.apps')
    @mock.patch('resolwe.flow.executors.settings')
    def test_not_list(self, settings_mock, apps_mock):
        apps_mock.get_app_configs.return_value = []
        settings_mock.RESOLWE_CUSTOM_TOOLS_PATHS = '/custom_tools'

        base_executor = BaseFlowExecutor()
        with six.assertRaisesRegex(self, KeyError, 'setting must be a list'):
            base_executor.get_tools()


class SpawnedProcessTest(ProcessTestCase):
    def setUp(self):
        super(SpawnedProcessTest, self).setUp()
        self._register_schemas(path=[PROCESSES_DIR])

    def test_test(self):
        self.run_process('test-spawn-new')

        data = Data.objects.last()
        data_dir = settings.FLOW_EXECUTOR['DATA_DIR']
        file_path = os.path.join(data_dir, str(data.pk), 'foo.bar')
        self.assertEqual(data.output['saved_file']['file'], 'foo.bar')
        self.assertTrue(os.path.isfile(file_path))



from __future__ import absolute_import

from celery import shared_task  # pylint: disable=import-error


@shared_task
def celery_run(executor, data_id, script, verbosity):
    executor.run(data_id, script, verbosity)



# pylint: disable=missing-docstring
from __future__ import absolute_import, division, print_function, unicode_literals

import os
import shutil

from django import template
from django.conf import settings
from django.contrib.auth import get_user_model
from django.test import TestCase

from resolwe.flow.managers import manager
from resolwe.flow.models import Process, Data


class ProcessFieldsTagsTest(TestCase):
    def setUp(self):
        User = get_user_model()
        self.contributor = User.objects.create_user('test_user', 'test_pwd')

    def tearDown(self):
        for data in Data.objects.all():
            data_dir = os.path.join(settings.FLOW_EXECUTOR['DATA_DIR'], str(data.id))
            shutil.rmtree(data_dir, ignore_errors=True)

    def test_templatetags(self):
        input_process = Process.objects.create(
            name='Input process',
            contributor=self.contributor,
            type='data:test:inputobject:',
        )
        input_data = Data.objects.create(
            name='Input Data object',
            contributor=self.contributor,
            process=input_process,
        )

        process = Process.objects.create(
            name='Test template tags',
            contributor=self.contributor,
            type='test:data:templatetags:',
            input_schema=[
                {'name': 'input_data', 'type': 'data:test:inputobject:'},
            ],
            output_schema=[
                {'name': 'name', 'type': 'basic:string:'},
                {'name': 'id', 'type': 'basic:integer:'},
                {'name': 'type', 'type': 'basic:string:'},
            ],
            run={
                'bash': """
re-save name "{{ input_data | name }}"
re-save id {{ input_data | id }}
re-save type {{ input_data | type }}
"""
            }

        )
        data = Data.objects.create(
            name='Data object',
            contributor=self.contributor,
            process=process,
            input={'input_data': input_data.pk},
        )

        manager.communicate(verbosity=0)

        # update output
        data = Data.objects.get(pk=data.pk)

        self.assertEqual(data.output['name'], input_data.name)
        self.assertEqual(data.output['id'], input_data.pk)
        self.assertEqual(data.output['type'], input_process.type)



"""
Django settings for running tests for Resolwe package.

"""
from __future__ import absolute_import, division, print_function, unicode_literals

import os

PROJECT_ROOT = os.path.abspath(os.path.dirname(__file__))

SECRET_KEY = 'secret'

DEBUG = True

MIDDLEWARE_CLASSES = (
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
)

# Apps from this project
PROJECT_APPS = (
    'resolwe',
    'resolwe.permissions',
    'resolwe.flow',
)

INSTALLED_APPS = (
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.staticfiles',

    # 'kombu.transport.django',  # required for Celery to work with Django DB.

    'rest_framework',
    'guardian',
    'mathfilters',
    'versionfield',
) + PROJECT_APPS

ROOT_URLCONF = 'tests.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
            ],
        },
    },
]

AUTHENTICATION_BACKENDS = (
    'django.contrib.auth.backends.ModelBackend',
    'guardian.backends.ObjectPermissionBackend',
)

ANONYMOUS_USER_NAME = 'public'

# Get the current Tox testing environment
# NOTE: This is useful for concurrently running tests with separate environments
toxenv = os.environ.get('TOXENV', '')

# Check if PostgreSQL settings are set via environment variables
pgname = os.environ.get('RESOLWE_POSTGRESQL_NAME', 'resolwe')
pguser = os.environ.get('RESOLWE_POSTGRESQL_USER', 'resolwe')
pghost = os.environ.get('RESOLWE_POSTGRESQL_HOST', 'localhost')
pgport = int(os.environ.get('RESOLWE_POSTGRESQL_PORT', 55432))

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.postgresql_psycopg2',
        'NAME': pgname,
        'USER': pguser,
        'HOST': pghost,
        'PORT': pgport,
        'TEST': {
            'NAME': 'resolwe_test' + toxenv
        }
    }
}

BROKER_URL = 'django://'
CELERY_TASK_SERIALIZER = 'json'
CELERY_ACCEPT_CONTENT = [CELERY_TASK_SERIALIZER]

STATIC_URL = '/static/'

FLOW_EXECUTOR = {
    'NAME': 'resolwe.flow.executors.local',
    'DATA_DIR': os.path.join(PROJECT_ROOT, '.test_data'),
    'UPLOAD_DIR': os.path.join(PROJECT_ROOT, '.test_upload'),
}
# Set custom executor command if set via environment variable
if 'RESOLWE_EXECUTOR_COMMAND' in os.environ:
    FLOW_EXECUTOR['COMMAND'] = os.environ['RESOLWE_EXECUTOR_COMMAND']
FLOW_API = {
    'PERMISSIONS': 'resolwe.permissions.permissions',
}
FLOW_EXPRESSION_ENGINES = [
    'resolwe.flow.exprengines.dtlbash'
]

FLOW_PROCESSES_FINDERS = (
    'resolwe.flow.finders.FileSystemProcessesFinder',
    'resolwe.flow.finders.AppDirectoriesFinder',
)

REST_FRAMEWORK = {
    'DEFAULT_AUTHENTICATION_CLASSES': (
        'rest_framework.authentication.SessionAuthentication',
    ),
    'DEFAULT_FILTER_BACKENDS': (
        'resolwe.permissions.filters.ResolwePermissionsFilter',
        'rest_framework_filters.backends.DjangoFilterBackend',
    ),
}



"""
===========================
Process Fields Templatetags
===========================

"""
from django import template

from resolwe.flow.models import Data


register = template.Library()  # pylint: disable=invalid-name


def _get_data_attr(data, attr):
    if isinstance(data, dict):
        # `Data` object's id is hydrated as `__id` in expression engine
        data = data['__id']

    data_obj = Data.objects.get(id=data)

    return getattr(data_obj, attr)


@register.filter
def name(data):
    """Return `name` of `Data`."""
    return _get_data_attr(data, 'name')


@register.filter
def id(obj):  # pylint: disable=redefined-builtin,invalid-name
    """Return ``id`` key of dict."""
    return obj['__id']


@register.filter
def type(obj):  # pylint: disable=redefined-builtin
    """Return ``type`` key of dict."""
    return obj['__type']



"""
Sphinx extension for semi-automatic documentation of Resolwe processes.

This module introduces a new directive that enables user to document a
set of processes. The directive can be used in three variations::

    .. autoprocess::

    .. autoprocess:: category_name

    .. autoprocess:: process_slug

The first option documents *all* processes.

The second option documents just the ones that have *category* equal to
``category_name``. This means that subcategories (e.g. ``analyses:alignment``,
``analyses:variants``) of a category (e.g. ``analyses``) will not be documented.

Third option documents only one process: the one with *slug* equal to
``process_slug``.
"""

import os
import re
import fnmatch

import six
import yaml

from docutils import nodes
from docutils.parsers.rst import Directive


class AutoProcessDirective(Directive):
    """
    Directive used to automatically document a Resolwe process.
    """
    has_content = True
    required_arguments = 0
    optional_arguments = 0
    final_argument_whitespace = False
    option_spec = None

    def filter_processes(self, content, categories, processes_dict):
        """
        Filter processes_dict by the content of the directive

        Content can be:

            * None - if there is no content, all processes are returned.
            * Category - if the content of the directive is a string matching
              an element in categories list, only processes with this
              category will be returned.
            * Process slug - if the content of the directive is a valid
              process slug, only the corresponding process is returned
              (packed in a list)

        :param list content: content of the directive
        :param set categories: set of all categories
        :param dict processes_dict: dict of all source_base_url: processes pairs
        :return: filtered processes_dict
        :rtype: dict
        :raises: NotImplementedError: if more than one argument is passed
        :raises: ValueError: if the given argument is ivalid
        """
        if not content:
            return processes_dict
        elif len(content) > 1:
            raise NotImplementedError("Only one argument can be passed to autoprocess.")
        content = content[0]
        # Make a filtered dict of all process that have slug == content
        filtered_by_slug = {url: proc for url, proc in six.iteritems(processes_dict) if proc['slug'] == content}
        if len(filtered_by_slug) == 1:
            return filtered_by_slug

        elif content in categories:
            filtered_by_category = {
                url: proc for url, proc in six.iteritems(processes_dict)
                if proc.get('category', 'uncategorized') == content}
            return filtered_by_category

        raise ValueError("Invalid argument given to autoprocess directive.")

    def get_process_definition_start(self, fname, slug):
        """
        Find the starting line of process definition - where slug is defined

        :param str fname: Path to filename with processes
        :param string slug: process slug
        :return: line where the process definiton starts
        :rtype: int
        """
        with open(fname) as file_:
            for i, line in enumerate(file_):
                if re.search(r'slug:\s*{}'.format(slug), line):
                    return i + 1
        # In case starting line is not found just return first line
        return 1

    def get_all_processes(self, process_dir, source_base_url):
        """
        :param str process_dir: Path to the directory where to search for processes
        :param str source_base_url: Base URL of the source code repository with process definitions
        :return: Dictionary of processes where keys are URLs pointing to processes'
        source code and values are processes' definitions parsed from YAML files
        :rtype: dict
        :raises: ValueError: if multiple processes with the same slug are found
        """
        all_process_files = []
        process_file_extensions = ['*.yaml', '*.yml']
        for root, _, filenames in os.walk(process_dir):
            for extension in process_file_extensions:
                for filename in fnmatch.filter(filenames, extension):
                    all_process_files.append(os.path.join(root, filename))

        def read_yaml_file(fname):
            """Read yaml file"""
            with open(fname) as file_:
                return yaml.load(file_)

        all_processes = {}
        for process_file in all_process_files:
            processes_in_file = read_yaml_file(process_file)
            for process in processes_in_file:

                # This section finds the line in file where the
                # defintion of the process starts. (there are
                # multiple process definition in some files).
                startline = self.get_process_definition_start(process_file, process['slug'])

                # Put together URL to starting line of process definition.
                source_url = source_base_url + os.path.basename(process_file) + '#L' + str(startline)
                chosen_process = [proc for proc in processes_in_file if proc['slug'] == process['slug']]
                # The scenarion with len(chosen_process) == 0 is not possible
                if len(chosen_process) > 1:
                    raise ValueError("More than one process with same slug: {}".format(process['slug']))

                all_processes[source_url] = chosen_process[0]

        return all_processes

    def get_all_categories(self, process_list):
        """
        Get list of all categories in given process list.

        For processes with undefined category field (yes, there are some!),
        there is "uncategorized" category.

        :param list process_list: list of all processes
        :return: list of all categories in given process list
        :rtype: list
        """
        categories = set()
        for process in process_list:
            categories.add(process.get('category', 'uncategorized'))
        return list(categories)

    def make_field(self, field_name, field_body):
        """
        Handy function for filling content into nodes.field

        :param string field_name: Field name of the field
        :param field_name: Field body if the field
        :type field_name: str or instance of docutils.nodes
        :return: field instance filled with given name and body
        :rtype: nodes.field
        """

        name = nodes.field_name()
        name += nodes.Text(field_name)

        paragraph = nodes.paragraph()
        if isinstance(field_body, six.string_types):
            # This is the case when field_body is just a string:
            paragraph += nodes.Text(field_body)
        else:
            # This is the case when field_body is a complex node:
            # useful when constructing nested field lists
            paragraph += field_body

        body = nodes.field_body()
        body += paragraph

        field = nodes.field()
        field.extend([name, body])
        return field

    def make_properties_list(self, data):
        """
        Fill the ``data`` into a properties list and return it.

        :param dict data: the content of the property list to make
        :return: field_list instance filled with given data
        :rtype: nodes.field_list
        """
        # Make a copy so the function parameter is not modified.
        data = data.copy()

        properties_list = nodes.field_list()

        # Remove the name (it is already used in field_name...)
        data.pop('name')

        # changing the order of elements in this list affects
        # the order in which they are displayed
        property_names = [
            'label',
            'type',
            'description',
            'required',
            'disabled',
            'hidden',
            'default',
            'placeholder',
            'validate_regex',
            'choices',
            'collapse',
            'group']

        for name in property_names:
            if name in data:
                value = data.pop(name)

                # Value should be formatted in code-style (=literal) mode
                if name in ['type', 'default', 'placeholder', 'validate_regex']:
                    literal_node = nodes.literal(str(value), str(value))
                    properties_list += self.make_field(name, literal_node)

                # Special formating of ``value`` is needed if name == 'choices'
                elif name == 'choices':
                    bullet_list = nodes.bullet_list()
                    for choice in value:
                        label = nodes.Text(choice['label'] + ': ')
                        val = nodes.literal(choice['value'], choice['value'])

                        paragraph = nodes.paragraph()
                        paragraph += label
                        paragraph += val
                        list_item = nodes.list_item()
                        list_item += paragraph
                        bullet_list += list_item

                    properties_list += self.make_field(name, bullet_list)

                # Make a nested list of properties:
                elif name == 'group':
                    list123 = nodes.field_list()
                    for sub_param in value:
                        sub_param_properties_list = self.make_properties_list(sub_param)
                        list123 += self.make_field(sub_param['name'], sub_param_properties_list)

                    properties_list += self.make_field(name, list123)

                else:
                    properties_list += self.make_field(name, str(value))

        return properties_list

    def make_process_node(self, source_base_url, process):
        """
        Fill the content of process definiton node

        :param str source_base_url: url to process definition
        :param dict process: process data as given from yaml.load function
        :return: section instance filled with given content
        :rtype: nodes.field
        """

        name = process['name']

        # Make process name as section title:
        section = nodes.section()
        section['ids'] = [process['slug']]
        title = nodes.subtitle()
        title_text = nodes.Text(name)
        title += title_text
        section += title

        top_field_list = nodes.field_list()

        slug = process['slug']
        version = process['version']
        description = process.get('description', '')
        inputs = process.get('input', [])
        outputs = process.get('output', [])

        # Make hyperlink to process definition:
        para = nodes.paragraph()
        source_reference = nodes.reference('Link', 'Link')
        source_reference['refdocname'] = 'some_random_string'
        source_reference['refuri'] = source_base_url
        para += source_reference

        # Make slug, varsion, description and source_code link:
        names = ['Slug', 'Version', 'Description', 'Source code']
        values = [slug, version, description, para]
        for name, value in zip(names, values):
            top_field_list += self.make_field(name, value)

        # Make inputs section:
        inputs_list = nodes.field_list()
        for input_ in inputs:
            input_properties_list = self.make_properties_list(input_)
            inputs_list += self.make_field(input_['name'], input_properties_list)
        top_field_list += self.make_field('Inputs', inputs_list)

        # Make outputs section:
        outputs_list = nodes.field_list()
        for output in outputs:
            output_properties_list = self.make_properties_list(output)
            outputs_list += self.make_field(output['name'], output_properties_list)
        top_field_list += self.make_field('Outputs', outputs_list)

        section += top_field_list
        return section

    def run(self):
        """
        Return a list of nodes that should be parsed.

        Classes derived from Directive class must implemenent run() method,
        which is the most important method of this class: it returns a list
        of nodes. Nodes are building blocks of the document and are the
        basic internal structure that is later formatted into HTML/LaTeX...

        Several things happen in run method:

            * the content of the directive will be parsed: this will
              determine which processes to document and extract the
              content of *.yaml files into dictionary like object.
              This is done by method ``filter_processes``
            * This dictionary-like object is traversed and desired
              information is extracted: name, slug, id, inputs, outputs...
            * With this information we cunstruct a list of nodes. Nodes are
              representing the hierarchical structure of the document.

        """
        config = self.state.document.settings.env.config

        # Get all processes:
        all_processes_dict = self.get_all_processes(config.autoprocess_process_dir, config.autoprocess_source_base_url)
        all_processes = [val for val in all_processes_dict.values()]

        # Extract all the categories:
        categories = self.get_all_categories(all_processes)

        # This is the list of processes thet need to be documented
        processes_to_document = self.filter_processes(self.content, categories, all_processes_dict)

        if len(processes_to_document) == 1:
            # Only one process:
            source_url, process = list(processes_to_document.items())[0]
            top_node = self.make_process_node(source_url, process)
        elif len(processes_to_document) > 1:
            # A category of processes, make a process category section:
            process1 = list(processes_to_document.items())[0][1]
            category = process1.get('category', 'uncategorized')
            title = nodes.title()
            title += nodes.Text("Category: " + str(category))
            top_node = nodes.section()
            top_node['ids'] = [category]
            top_node += title

            # Make sorting:
            pairs = sorted(processes_to_document.items(), key=lambda pair: pair[1]['name'])
            for source_url, process in pairs:
                process_node = self.make_process_node(source_url, process)
                top_node += process_node
        else:
            raise ValueError("No processes to documement.")

        return [top_node]


def setup(app):
    """
    When sphinx loads the extension (= imports the extension module) it
    also executes the setup() function. Setup is the way extension
    informs Sphinx about everything that the extension enables: which
    config_values are introduced, which custom nodes/directives/roles
    and which events are defined in extension.

    In this case, only one new directive is created. All used nodes are
    constructed from already existing nodes in docutils.nodes package.
    """
    app.add_config_value('autoprocess_process_dir', '', 'env')
    app.add_config_value('autoprocess_source_base_url', '', 'env')

    app.add_directive('autoprocess', AutoProcessDirective)

    # The setup() function can return a dictionary. This is treated by
    # Sphinx as metadata of the extension:
    return {'version': '0.1'}



# pylint: disable=missing-docstring
from __future__ import absolute_import, division, print_function, unicode_literals

from django.core.urlresolvers import reverse
from django.test import TestCase

from rest_framework.test import APIRequestFactory


class CollectionResourceTest(TestCase):
    def setUp(self):
        self.api_client = APIRequestFactory()
        self.list_url = reverse('resolwe-api:collection-list')
        # TODO: add object's id
        # self.detail_url = reverse('resolwe-api:collection-detail')

    def test_readonly_fields(self):
        self.api_client.get(self.list_url)



"""
==============
Flow Utilities
==============

.. automodule:: resolwe.flow.utils.purge
   :members:
.. automodule:: resolwe.flow.utils.test
   :members:

"""
from __future__ import absolute_import, division, print_function, unicode_literals



# -*- coding: utf-8 -*-
# pylint: disable=missing-docstring
from __future__ import absolute_import, division, print_function, unicode_literals

import gzip
import io
import json
import mock
import os.path
import sys
import tempfile
import unittest

try:
    import builtins  # py3
except ImportError:
    import __builtin__ as builtins  # py2

import six

from resolwe.flow.models import Storage
from resolwe.flow.utils.test import ProcessTestCase


class TestingFrameworkTestCase(unittest.TestCase):

        @mock.patch("os.path.isfile")
        @mock.patch("resolwe.flow.utils.test.dict_dot")
        def test_assert_files_date_not_filtered(self, dict_dot_mock, isfile_mock):
            isfile_mock.return_value = True
            output1_file = io.BytesIO(b"some line\ndate: 2016-02-10\n")
            output2_file = io.BytesIO(b"some line\ndate: 2015-10-31\n")
            open_mock = mock.MagicMock(side_effect=[output1_file, output2_file])
            with mock.patch.object(builtins, 'open', open_mock):
                dummy_case = ProcessTestCase.__new__(ProcessTestCase)
                dummy_case.files_path = ""
                dummy_case._debug_info = lambda _: ""
                dummy_case.assertEqual = self.assertEqual
                obj_mock = mock.MagicMock()
                self.assertRaises(AssertionError, ProcessTestCase._assert_file,
                                  dummy_case, obj_mock, "", "")

        @mock.patch("os.path.isfile")
        @mock.patch("resolwe.flow.utils.test.dict_dot")
        def test_assert_files_date_filtered(self, dict_dot_mock, isfile_mock):
            isfile_mock.return_value = True
            output1_file = io.BytesIO(b"some line\ndate: 2016-02-10\n")
            output2_file = io.BytesIO(b"some line\ndate: 2015-10-31\n")
            open_mock = mock.MagicMock(side_effect=[output1_file, output2_file])
            with mock.patch.object(builtins, 'open', open_mock):
                dummy_case = ProcessTestCase.__new__(ProcessTestCase)
                dummy_case.files_path = ""
                dummy_case._debug_info = lambda _: ""
                dummy_case.assertEqual = self.assertEqual
                obj_mock = mock.MagicMock()

                def date_in_line(x):
                    return x.startswith(b"date")

                ProcessTestCase._assert_file(dummy_case, obj_mock, "", "", filter=date_in_line)

        @unittest.skipIf(six.PY2, "Needs Python 3 version of the gzip module")
        def test_assert_json_storage_object(self):
            example_json = {'foo': [1.0, 2.5, 3.14], 'bar': ['ba', 'cd']}

            dummy_case = ProcessTestCase.__new__(ProcessTestCase)
            dummy_case.files_path = ''
            dummy_case._debug_info = lambda _: ''
            dummy_case.assertEqual = self.assertEqual

            obj_mock = mock.MagicMock()

            storage_mock = mock.MagicMock(spec=Storage)
            storage_mock.id = 'no_id'
            storage_mock.json = example_json

            # use in-memory binary stream object for speed and simplicity
            gzipped_json_file = io.BytesIO()
            with gzip.open(gzipped_json_file, mode='wt') as f:
                json.dump(example_json, f)
            # set seek position of the binary stream object back to 0
            gzipped_json_file.seek(0)

            join_mock = mock.MagicMock(side_effect=[gzipped_json_file])
            isfile_mock = mock.MagicMock(return_value=True)
            with mock.patch.object(os.path, 'join', join_mock):
                with mock.patch.object(os.path, 'isfile', isfile_mock):
                    ProcessTestCase.assertJSON(dummy_case, obj_mock, storage_mock, '', 'foo.gz')

        @unittest.skipIf(six.PY2, "Needs Python 3 version of the gzip module")
        @mock.patch('resolwe.flow.models.Storage.objects.get')
        def test_assert_json_storage_id(self, get_mock):
            example_json = {'foo': [1.0, 2.5, 3.14], 'bar': ['ba', 'cd']}

            dummy_case = ProcessTestCase.__new__(ProcessTestCase)
            dummy_case.files_path = ''
            dummy_case._debug_info = lambda _: ''
            dummy_case.assertEqual = self.assertEqual

            obj_mock = mock.MagicMock()

            storage_id = 'no_id'
            storage_mock = mock.MagicMock(spec=Storage)
            storage_mock.id = storage_id
            storage_mock.json = example_json
            get_mock.side_effect = [storage_mock]

            # use in-memory binary stream object for speed and simplicity
            gzipped_json_file = io.BytesIO()
            with gzip.open(gzipped_json_file, mode='wt') as f:
                json.dump(example_json, f)
            # set seek position of the binary stream object back to 0
            gzipped_json_file.seek(0)

            join_mock = mock.MagicMock(side_effect=[gzipped_json_file])
            isfile_mock = mock.MagicMock(return_value=True)
            with mock.patch.object(os.path, 'join', join_mock):
                with mock.patch.object(os.path, 'isfile', isfile_mock):
                    ProcessTestCase.assertJSON(dummy_case, obj_mock, storage_id, '', 'foo.gz')

        @unittest.skipIf(six.PY2, "Needs Python 3 version of the gzip module")
        def test_assert_json_file_missing(self):
            example_json = {'foo': [1.0, 2.5, 3.14], 'bar': ['ba', 'cd']}

            dummy_case = ProcessTestCase.__new__(ProcessTestCase)
            dummy_case.files_path = ''
            dummy_case._debug_info = lambda _: ''
            dummy_case.assertEqual = self.assertEqual

            obj_mock = mock.MagicMock()

            storage_mock = mock.MagicMock(spec=Storage)
            storage_mock.id = 'no_id'
            storage_mock.json = example_json

            # use in-memory binary stream object for speed and simplicity
            gzipped_json_file = io.BytesIO()
            join_modified_values = [gzipped_json_file]
            orig_join = os.path.join

            # NOTE: coverage tool needs original 'os.path.join'
            def join_side_effect(path, *paths):
                if len(join_modified_values) > 0:
                    return join_modified_values.pop(0)
                else:
                    return orig_join(path, *paths)

            join_mock = mock.MagicMock(side_effect=join_side_effect)
            isfile_mock = mock.MagicMock(return_value=False)
            with mock.patch.object(os.path, 'join', join_mock):
                with mock.patch.object(os.path, 'isfile', isfile_mock):
                    with six.assertRaisesRegex(self, AssertionError,
                                               'Output file .* missing so it was created.'):
                        ProcessTestCase.assertJSON(dummy_case, obj_mock, storage_mock, '', 'foo.gz')

            # set seek position of the binary stream object back to 0
            gzipped_json_file.seek(0)
            with gzip.open(gzipped_json_file, mode='rt') as f:
                unzipped_json = json.load(f)
            self.assertEqual(example_json, unzipped_json)

        def test_debug_info_non_ascii_stdout_file(self):

            non_ascii_text = 'Some non-ascii chars č ü €'

            dummy_case = ProcessTestCase.__new__(ProcessTestCase)

            obj_mock = mock.MagicMock()
            obj_mock.pk = 'no_id'

            with tempfile.NamedTemporaryFile(mode='wb') as stdout_file:
                stdout_file.write(non_ascii_text.encode('utf-8'))
                # set seek position of the file back to 0
                stdout_file.seek(0)
                with mock.patch.object(os.path, 'join', side_effect=[stdout_file.name]):
                    with mock.patch.object(os.path, 'isfile', return_value=True):
                        six.assertRegex(self, ProcessTestCase._debug_info(dummy_case, obj_mock),
                                        non_ascii_text)



"""
============
Flow Filters
============

"""
from __future__ import absolute_import, division, print_function, unicode_literals

import rest_framework_filters as filters

from .models import Collection, Data, DescriptorSchema, Process


class DescriptorSchemaFilter(filters.FilterSet):
    class Meta:
        model = DescriptorSchema
        fields = {
            'slug': '__all__',
            'id': '__all__',
        }


class CollectionFilter(filters.FilterSet):
    data = filters.ModelChoiceFilter(queryset=Data.objects.all())
    descriptor_schema = filters.RelatedFilter(DescriptorSchemaFilter, name='descriptor_schema')

    class Meta:
        model = Collection
        fields = {
            'contributor': ['exact', ],
            'created': '__all__',
            'data': ['exact', ],
            'description': '__all__',
            'descriptor_schema': ['exact', ],
            'id': '__all__',
            'modified': '__all__',
            'name': '__all__',
            'slug': '__all__',
        }


class DataFilter(filters.FilterSet):
    collection = filters.ModelChoiceFilter(queryset=Collection.objects.all())
    type = filters.CharFilter(name='process__type', lookup_type='startswith')

    class Meta:
        model = Data
        fields = {
            'collection': ['exact', ],
            'contributor': ['exact', ],
            'created': '__all__',
            'finished': '__all__',
            'id': '__all__',
            'modified': '__all__',
            'name': '__all__',
            'process': ['exact', ],
            'slug': '__all__',
            'started': '__all__',
            'status': '__all__',
            'type': ['exact', ],
        }


class ProcessFilter(filters.FilterSet):
    category = filters.CharFilter(name='category', lookup_type='startswith')

    class Meta:
        model = Process
        fields = {
            'category': ['exact', ],
            'contributor': ['exact', ],
            'created': '__all__',
            'id': '__all__',
            'modified': '__all__',
            'name': '__all__',
            'slug': '__all__',
        }



from django.db.models.signals import post_save
from django.dispatch import receiver

from resolwe.flow.models import Data
from resolwe.flow.managers import manager


@receiver(post_save, sender=Data)
def manager_post_save_handler(sender, instance, **kwargs):
    if instance.status == Data.STATUS_DONE or instance.status == Data.STATUS_ERROR:
        manager.communicate(verbosity=0)



"""
===========
Flow Models
===========

Base Model
==========

Base model for all other models.

.. autoclass:: resolwe.flow.models.BaseModel
    :members:


Collection Model
================

Postgres ORM model for the organization of collections.

.. autoclass:: resolwe.flow.models.BaseCollection
    :members:

.. autoclass:: resolwe.flow.models.Collection
    :members:


Data model
==========

Postgres ORM model for keeping the data structured.

.. autoclass:: resolwe.flow.models.Data
    :members:


DescriptorSchema model
======================

Postgres ORM model for storing descriptors.

.. autoclass:: resolwe.flow.models.DescriptorSchema
    :members:


Process model
=============

Postgress ORM model for storing processes.

.. autoclass:: resolwe.flow.models.Process
    :members:


Storage model
=============

Postgres ORM model for storing JSON.

.. autoclass:: resolwe.flow.models.Storage
    :members:

"""
from __future__ import absolute_import, division, print_function, unicode_literals
import functools
import json
import jsonschema
import os
import re
import six

from django import template
from django.db import models, transaction
from django.conf import settings
from django.core.exceptions import ValidationError
from django.core.validators import RegexValidator
from django.contrib.postgres.fields import ArrayField, JSONField
from django.contrib.staticfiles import finders

from versionfield import VersionField
from autoslug import AutoSlugField


VERSION_NUMBER_BITS = (8, 10, 14)


class BaseModel(models.Model):

    """Abstract model that includes common fields for other models."""

    class Meta:
        """BaseModel Meta options."""
        abstract = True
        unique_together = ('slug', 'version')
        default_permissions = ()

    #: URL slug
    slug = AutoSlugField(populate_from='name', unique_with='version', editable=True, max_length=100)

    #: process version
    version = VersionField(number_bits=VERSION_NUMBER_BITS, default=0)

    #: object name
    name = models.CharField(max_length=100)

    #: creation date and time
    created = models.DateTimeField(auto_now_add=True)

    #: modified date and time
    modified = models.DateTimeField(auto_now=True)

    #: user that created the entry
    contributor = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.PROTECT)

    def __str__(self):
        return self.name


class Process(BaseModel):

    """Postgres model for storing processes."""

    class Meta(BaseModel.Meta):
        """Process Meta options."""
        permissions = (
            ("view_process", "Can view process"),
            ("share_process", "Can share process"),
            ("owner_process", "Is owner of the process"),
        )

    #: raw persistence
    PERSISTENCE_RAW = 'RAW'
    #: cached persistence
    PERSISTENCE_CACHED = 'CAC'
    #: temp persistence
    PERSISTENCE_TEMP = 'TMP'
    PERSISTENCE_CHOICES = (
        (PERSISTENCE_RAW, 'Raw'),
        (PERSISTENCE_CACHED, 'Cached'),
        (PERSISTENCE_TEMP, 'Temp'),
    )

    #: high priority
    PRIORITY_HIGH = 'HI'
    #: normal priority
    PRIORITY_NORMAL = 'NO'
    PRIORITY_CHOICES = (
        (PRIORITY_NORMAL, 'Normal'),
        (PRIORITY_HIGH, 'High'),
    )

    #: data type
    type = models.CharField(max_length=100, validators=[
        RegexValidator(
            regex=r'^data:[a-z0-9:]+:$',
            message='Type may be alphanumerics separated by colon',
            code='invalid_type'
        )
    ])

    #: category
    category = models.CharField(max_length=200, default='other', validators=[
        RegexValidator(
            regex=r'^([a-z0-9]+[:\-])*[a-z0-9]+:$',
            message='Category may be alphanumerics separated by colon',
            code='invalid_category'
        )
    ])

    persistence = models.CharField(max_length=3, choices=PERSISTENCE_CHOICES, default=PERSISTENCE_RAW)
    """
    Persistence of :class:`~resolwe.flow.models.Data` objects created
    with this process. It can be one of the following:

    - :attr:`PERSISTENCE_RAW`
    - :attr:`PERSISTENCE_CACHED`
    - :attr:`PERSISTENCE_TEMP`

    .. note::

        If persistence is set to ``PERSISTENCE_CACHED`` or
        ``PERSISTENCE_TEMP``, the process must be idempotent.

    """

    priority = models.CharField(max_length=2, choices=PRIORITY_CHOICES, default=PRIORITY_NORMAL)
    """
    Process' execution priority. It can be one of the following:

    - :attr:`PRIORITY_NORMAL`
    - :attr:`PRIORITY_HIGH`

    """

    #: detailed description
    description = models.TextField(default='')

    #: template for name of Data object created with Process
    data_name = models.CharField(max_length=200, null=True, blank=True)

    input_schema = JSONField(blank=True, default=list)
    """
    process input schema (describes input parameters, form layout **"Inputs"** for :attr:`Data.input`)

    Handling:

    - schema defined by: *dev*
    - default by: *user*
    - changable by: *none*

    """

    output_schema = JSONField(blank=True, default=list)
    """
    process output schema (describes output JSON, form layout **"Results"** for :attr:`Data.output`)

    Handling:

    - schema defined by: *dev*
    - default by: *dev*
    - changable by: *dev*

    Implicitly defined fields (by
    :func:`resolwe.flow.management.commands.register` or
    :meth:`resolwe.flow.executors.BaseFlowExecutor.run` or its
    derivatives):

    - ``progress`` of type ``basic:float`` (from 0.0 to 1.0)
    - ``proc`` of type ``basic:group`` containing:

      - ``stdout`` of type ``basic:text``
      - ``rc`` of type ``basic:integer``
      - ``task`` of type ``basic:string`` (celery task id)
      - ``worker`` of type ``basic:string`` (celery worker hostname)
      - ``runtime`` of type ``basic:string`` (runtime instance hostname)
      - ``pid`` of type ``basic:integer`` (process ID)

    """

    flow_collection = models.CharField(max_length=100, null=True, blank=True)
    """
    Automatically add :class:`~resolwe.flow.models.Data` object created
    with this process to a special
    :class:`~resolwe.flow.models.Collection` object representing a
    data-flow. If all input ``Data`` objects belong to the same
    collection, add newly created ``Data`` object to it, otherwise
    create a new collection.
    If :class:`~resolwe.flow.models.DescriptorSchema` object with
    ``type`` matching this field exists, reference it in the
    collection's
    :attr:`~resolwe.flow.models.BaseCollection.descriptor_schema`
    field.

    """

    run = JSONField(default=dict)
    """
    process command and environment description for internal use

    Handling:

    - schema defined by: *dev*
    - default by: *dev*
    - changable by: *dev*

    Required definitions:

    - ``engine`` .. engine to run the processor with
    - ``script`` .. script with code to run

    """


def render_template(template_string, context):
    """Render template based on Dango template language."""

    template_headers = [
        '{% load resource_filters %}',
        '{% load process_fields %}',
        '{% load mathfilters %}',
    ]

    custom_template_tags = getattr(settings, 'RESOLWE_CUSTOM_TEMPLATE_TAGS', [])
    if not isinstance(custom_template_tags, list):
        raise KeyError("`RESOLWE_CUSTOM_TEMPLATE_TAGS` setting must be a list.")

    template_headers.extend(
        ['{{% load {} %}}'.format(template_tag) for template_tag in custom_template_tags]
    )

    return template.Template(''.join(template_headers) + template_string).render(context)


def render_descriptor(data):
    """Render data descriptor.

    The rendering is based on descriptor schema and input context.

    :param data: data instance
    :type data: :class:`resolwe.flow.models.Data` or :class:`dict`

    """
    if not data.descriptor_schema or not data.process.input_schema:
        return

    inputs = data.input.copy()
    hydrate_input_references(inputs, data.process.input_schema, hydrate_values=False)
    template_context = template.Context(inputs)

    # Set default values
    for field_schema, _, path in iterate_schema(data.descriptor, data.descriptor_schema.schema, 'descriptor'):
        if 'default' in field_schema:
            tmpl = field_schema['default']
            if field_schema['type'].startswith('list:'):
                tmpl = [render_template(tmp, template_context)
                        if isinstance(tmp, six.string_types) else tmp
                        for tmp in tmpl]
            elif isinstance(tmpl, six.string_types):
                tmpl = render_template(tmpl, template_context)

            dict_dot(data, path, tmpl)


class Data(BaseModel):

    """Postgres model for storing data."""

    class Meta(BaseModel.Meta):
        """Data Meta options."""
        permissions = (
            ("view_data", "Can view data"),
            ("edit_data", "Can edit data"),
            ("share_data", "Can share data"),
            ("download_data", "Can download files from data"),
            ("owner_data", "Is owner of the data"),
        )

    #: data object is uploading
    STATUS_UPLOADING = 'UP'
    #: data object is being resolved
    STATUS_RESOLVING = 'RE'
    #: data object is waiting
    STATUS_WAITING = 'WT'
    #: data object is processing
    STATUS_PROCESSING = 'PR'
    #: data object is done
    STATUS_DONE = 'OK'
    #: data object is in error state
    STATUS_ERROR = 'ER'
    #: data object is in dirty state
    STATUS_DIRTY = 'DR'
    STATUS_CHOICES = (
        (STATUS_UPLOADING, 'Uploading'),
        (STATUS_RESOLVING, 'Resolving'),
        (STATUS_WAITING, 'Waiting'),
        (STATUS_PROCESSING, 'Processing'),
        (STATUS_DONE, 'Done'),
        (STATUS_ERROR, 'Error'),
        (STATUS_DIRTY, 'Dirty')
    )

    #: process started date and time (set by
    #: :meth:`resolwe.flow.executors.BaseFlowExecutor.run` or its derivatives)
    started = models.DateTimeField(blank=True, null=True)

    #: process finished date date and time (set by
    #: :meth:`resolwe.flow.executors.BaseFlowExecutor.run` or its derivatives)
    finished = models.DateTimeField(blank=True, null=True)

    #: checksum field calculated on inputs
    checksum = models.CharField(max_length=40, validators=[
        RegexValidator(
            regex=r'^[0-9a-f]{40}$',
            message='Checksum is exactly 40 alphanumerics',
            code='invalid_checksum'
        )
    ], blank=True, null=True)

    status = models.CharField(max_length=2, choices=STATUS_CHOICES, default=STATUS_RESOLVING)
    """
    :class:`Data` status

    It can be one of the following:

    - :attr:`STATUS_UPLOADING`
    - :attr:`STATUS_RESOLVING`
    - :attr:`STATUS_WAITING`
    - :attr:`STATUS_PROCESSING`
    - :attr:`STATUS_DONE`
    - :attr:`STATUS_ERROR`
    """

    #: process used to compute the data object
    process = models.ForeignKey('Process', on_delete=models.PROTECT)

    #: process id
    process_pid = models.PositiveIntegerField(blank=True, null=True)

    #: progress
    process_progress = models.PositiveSmallIntegerField(default=0)

    #: return code
    process_rc = models.PositiveSmallIntegerField(blank=True, null=True)

    #: info log message
    process_info = ArrayField(models.CharField(max_length=255), default=[])

    #: warning log message
    process_warning = ArrayField(models.CharField(max_length=255), default=[])

    #: error log message
    process_error = ArrayField(models.CharField(max_length=255), default=[])

    #: actual inputs used by the process
    input = JSONField(default=dict)

    #: actual outputs of the process
    output = JSONField(default=dict)

    #: data descriptor schema
    descriptor_schema = models.ForeignKey('DescriptorSchema', blank=True, null=True, on_delete=models.PROTECT)

    #: actual descriptor
    descriptor = JSONField(default=dict)

    #: track if user set the data name explicitly
    named_by_user = models.BooleanField(default=False)

    #: dependencies between data objects
    parents = models.ManyToManyField(
        'self',
        symmetrical=False, related_name='children'
    )

    def __init__(self, *args, **kwargs):
        super(Data, self).__init__(*args, **kwargs)
        self._original_name = self.name

    def save_storage(self, instance, schema):
        """Save basic:json values to a Storage collection."""

        for field_schema, fields in iterate_fields(instance, schema):
            name = field_schema['name']
            value = fields[name]
            if field_schema.get('type', '').startswith('basic:json:'):
                if value and not self.pk:
                    raise ValidationError(
                        'Data object must be `created` before creating `basic:json:` fields')

                if isinstance(value, int):
                    # already in Storage
                    continue

                storage = Storage.objects.create(
                    name='Storage for data id {}'.format(self.pk),
                    contributor=self.contributor,
                    data_id=self.pk,
                    json=value,
                )

                # `value` is copied by value, so `fields[name]` must be changed
                fields[name] = storage.pk

    def save_dependencies(self, instance, schema):
        """Save data: and list:data: references as parents."""

        def add_dependency(value):
            try:
                self.parents.add(Data.objects.get(pk=value))  # pylint: disable=no-member
            except Data.DoesNotExist:
                pass

        for field_schema, fields in iterate_fields(instance, schema):
            name = field_schema['name']
            value = fields[name]

            if field_schema.get('type', '').startswith('data:'):
                add_dependency(value)
            elif field_schema.get('type', '').startswith('list:data:'):
                for data in value:
                    add_dependency(data)

    def save(self, render_name=False, *args, **kwargs):
        # Generate the descriptor if one is not already set.
        if self.name != self._original_name:
            self.named_by_user = True

        create = self.pk is None
        if create:
            # Default values for INPUT
            input_schema = self.process.input_schema  # pylint: disable=no-member
            for field_schema, fields, path in iterate_schema(self.input, input_schema):
                if 'default' in field_schema and field_schema['name'] not in fields:
                    dict_dot(self.input, path, field_schema['default'])

            if not self.name:
                self._render_name()
            else:
                self.named_by_user = True

        elif render_name:
            self._render_name()

        if not self.descriptor:
            render_descriptor(self)

        self.save_storage(self.output, self.process.output_schema)  # pylint: disable=no-member

        hydrate_size(self)

        if create:
            validate_schema(self.input, self.process.input_schema)  # pylint: disable=no-member

        if self.descriptor_schema:
            validate_schema(self.descriptor, self.descriptor_schema.schema)  # pylint: disable=no-member
        elif self.descriptor and self.descriptor != {}:
            raise ValueError("`descriptor_schema` must be defined if `descriptor` is given")

        path_prefix = os.path.join(settings.FLOW_EXECUTOR['DATA_DIR'], str(self.pk))
        output_schema = self.process.output_schema  # pylint: disable=no-member
        if self.status == Data.STATUS_DONE:
            validate_schema(self.output, output_schema, path_prefix=path_prefix)
        else:
            validate_schema(self.output, output_schema, path_prefix=path_prefix,
                            test_required=False)

        with transaction.atomic():
            super(Data, self).save(*args, **kwargs)

            # We can only save dependencies after the data object has been saved. This
            # is why a transaction block is needed and the save method must be called first.
            self.save_dependencies(self.input, self.process.input_schema)  # pylint: disable=no-member

    def _render_name(self):
        """Render data name.

        The rendering is based on name template (`process.data_name`) and
        input context.

        """
        if not self.process.data_name or self.named_by_user:  # pylint: disable=no-member
            return

        inputs = self.input.copy()  # pylint: disable=no-member
        hydrate_input_references(inputs, self.process.input_schema, hydrate_values=False)  # pylint: disable=no-member
        template_context = template.Context(inputs)

        self.name = render_template(self.process.data_name, template_context)  # pylint: disable=no-member


class DescriptorSchema(BaseModel):

    """Postgres model for storing descriptors."""

    class Meta(BaseModel.Meta):
        """DescriptorSchema Meta options."""
        permissions = (
            ("view_descriptorschema", "Can view descriptor schema"),
            ("edit_descriptorschema", "Can edit descriptor schema"),
            ("share_descriptorschema", "Can share descriptor schema"),
            ("owner_descriptorschema", "Is owner of the description schema"),
        )

    #: detailed description
    description = models.TextField(blank=True)

    #: user descriptor schema represented as a JSON object
    schema = JSONField(default=dict)


class Trigger(BaseModel):

    """Postgres model for storing triggers."""

    class Meta(BaseModel.Meta):
        """Data Meta options."""
        permissions = (
            ("view_trigger", "Can view trigger"),
            ("edit_trigger", "Can edit trigger"),
            ("share_trigger", "Can share trigger"),
            ("owner_trigger", "Is owner of the trigger"),
        )

    #: data type of triggering data objects
    type = models.CharField(max_length=100, validators=[
        RegexValidator(
            regex=r'^data:[a-z0-9:]+:$',
            message='Type may be alphanumerics separated by colon',
            code='invalid_type'
        )
    ])

    #: trigger condition
    trigger = models.CharField(max_length=500)

    #: path to where the id is inserted
    trigger_input = models.CharField(max_length=100)

    #: process used
    process = models.ForeignKey('Process', blank=True, null=True, on_delete=models.SET_NULL)

    #: input settings of the processor
    input = JSONField(default=dict)

    #: corresponding collection
    collection = models.ForeignKey('Collection')

    #: does the trigger run on its own
    autorun = models.BooleanField(default=False)


class Storage(BaseModel):

    """Postgres model for storing storages."""

    #: corresponding data object
    data = models.ForeignKey('Data')

    #: actual JSON stored
    json = JSONField()


class LazyStorageJSON(object):

    """Lazy load `json` attribute of `Storage` object."""

    def __init__(self, **kwargs):
        self._kwargs = kwargs
        self._json = None

    def _get_storage(self):
        """Load `json` field from `Storage` object."""
        if self._json is None:
            self._json = Storage.objects.get(**self._kwargs).json

    def __getitem__(self, key):
        self._get_storage()
        return self._json[key]

    def __repr__(self):
        self._get_storage()
        return self._json.__repr__()


class BaseCollection(BaseModel):

    """Template for Postgres model for storing a collection."""

    class Meta(BaseModel.Meta):
        """Collection Meta options."""
        abstract = True
        permissions = (
            ("view_collection", "Can view collection"),
            ("edit_collection", "Can edit collection"),
            ("share_collection", "Can share collection"),
            ("download_collection", "Can download files from collection"),
            ("add_collection", "Can add data objects to collection"),
            ("owner_collection", "Is owner of the collection"),
        )

    #: detailed description
    description = models.TextField(blank=True)

    settings = JSONField(default=dict)

    public_processes = models.ManyToManyField(Process)

    data = models.ManyToManyField(Data)

    #: collection descriptor schema
    descriptor_schema = models.ForeignKey(DescriptorSchema, blank=True, null=True, on_delete=models.PROTECT)

    #: collection descriptor
    descriptor = JSONField(default=dict)


class Collection(BaseCollection):

    """Postgres model for storing a collection."""

    pass


def iterate_fields(fields, schema, path_prefix=None):
    """Iterate over all field values sub-fields.

    This will iterate over all field values. Some fields defined in the schema
    might not be visited.

    :param fields: field values to iterate over
    :type fields: dict
    :param schema: schema to iterate over
    :type schema: dict
    :return: (field schema, field value)
    :rtype: tuple

    """
    if path_prefix is not None and path_prefix != '' and path_prefix[-1] != '.':
        path_prefix += '.'

    schema_dict = {val['name']: val for val in schema}
    for field_id, properties in fields.items():
        path = '{}{}'.format(path_prefix, field_id) if path_prefix is not None else None
        if field_id not in schema_dict:
            raise KeyError("Field definition ({}) missing in schema".format(field_id))
        if 'group' in schema_dict[field_id]:
            for rvals in iterate_fields(properties, schema_dict[field_id]['group'], path):
                yield (rvals if path_prefix is not None else rvals[:2])
        else:
            rvals = (schema_dict[field_id], fields, path)
            yield (rvals if path_prefix is not None else rvals[:2])


def iterate_schema(fields, schema, path_prefix=''):
    """Iterate over all schema sub-fields.

    This will iterate over all field definitions in the schema. Some field v
    alues might be None.

    :param fields: field values to iterate over
    :type fields: dict
    :param schema: schema to iterate over
    :type schema: dict
    :param path_prefix: dot separated path prefix
    :type path_prefix: str
    :return: (field schema, field value, field path)
    :rtype: tuple

    """
    if path_prefix and path_prefix[-1] != '.':
        path_prefix += '.'

    for field_schema in schema:
        name = field_schema['name']
        if 'group' in field_schema:
            for rvals in iterate_schema(fields[name] if name in fields else {},
                                        field_schema['group'], '{}{}'.format(path_prefix, name)):
                yield rvals
        else:
            yield (field_schema, fields, '{}{}'.format(path_prefix, name))


def validation_schema(name):
    """Return json schema for json validation."""
    schemas = {
        'processor': 'processorSchema.json',
        'descriptor': 'descriptorSchema.json',
        'field': 'fieldSchema.json',
        'type': 'typeSchema.json',
    }

    if name not in schemas:
        raise ValueError()

    field_schema_file = finders.find('flow/{}'.format(schemas['field']), all=True)[0]
    field_schema = open(field_schema_file, 'r').read()

    if name == 'field':
        return json.loads(field_schema.replace('{{PARENT}}', ''))

    schema_file = finders.find('flow/{}'.format(schemas[name]), all=True)[0]
    schema = open(schema_file, 'r').read()

    return json.loads(schema.replace('{{FIELD}}', field_schema).replace('{{PARENT}}', '/field'))


TYPE_SCHEMA = validation_schema('type')


def validate_schema(instance, schema, test_required=True, path_prefix=None):
    """Check if DictField values are consistent with our data types.

    Perform basic JSON schema validation and our custom validations:

      * check that required fields are given (if `test_required` is set
        to ``True``)
      * check if ``basic:file:`` and ``list:basic:file`` fields match
        regex given in schema (only if ``validate_regex`` is defined in
        schema for coresponding fields) and exists (only if
        ``path_prefix`` is given)
      * check if directories referenced in ``basic:dir:`` and
        ``list:basic:dir``fields exist (only if ``path_prefix`` is
        given)
      * check that referenced ``Data`` objects (in ``data:<data_type>``
        and  ``list:data:<data_type>`` fields) exists and are of type
        ``<data_type>``
      * check that referenced ``Storage`` objects (in ``basic:json``
        fields) exists

    :param list instance: Instance to be validated
    :param list schema: Schema for validation
    :param bool test_required: Flag for testing if all required fields
        are present. It is usefule if validation is run before ``Data``
        object is finished and there are some field stil missing
        (default: ``False``)
    :param str path_prefix: path prefix used for checking if files and
        directories exist (default: ``None``)
    :rtype: None
    :raises ValidationError: if ``instance`` doesn't match schema
        defined in ``schema``

    """
    def validate_refs(field):
        if 'refs' in field:
            for refs_filename in field['refs']:
                refs_path = os.path.join(path_prefix, refs_filename)
                if not (os.path.isfile(refs_path) or os.path.isdir(refs_path)):
                    raise ValidationError(
                        "File referenced in `refs` ({}) does not exist".format(refs_path))

    def validate_file(field, regex):
        """Validate file name (and check that it exists)."""
        filename = field['file']

        if regex and not re.search(regex, filename):
            raise ValidationError(
                "File name {} does not match regex {}".format(filename, regex))

        if path_prefix:
            path = os.path.join(path_prefix, filename)
            if not os.path.isfile(path):
                raise ValidationError("Referenced file ({}) does not exist".format(path))

            validate_refs(field)

    def validate_dir(field):
        """Check that dirs and referenced files exists."""
        dirname = field['dir']

        if path_prefix:
            path = os.path.join(path_prefix, dirname)
            if not os.path.isdir(path):
                raise ValidationError("Referenced dir ({}) does not exist".format(path))

            validate_refs(field)

    def validate_data(data_pk, type_):
        """"Check that `Data` objects exist and is of right type."""
        data_qs = Data.objects.filter(pk=data_pk).values('process__type')
        if not data_qs.exists():
            raise ValidationError(
                "Referenced `Data` object does not exist (id:{})".format(data_pk))
        data = data_qs.first()
        if not data['process__type'].startswith(type_):
            raise ValidationError(
                "Data object of type `{}` is required, but type `{}` is given. "
                "(id:{})".format(type_, data['process__type'], data_pk))

    for _schema, _fields, _ in iterate_schema(instance, schema):
        name = _schema['name']

        if test_required and _schema.get('required', True) and name not in _fields:
            raise ValidationError("Required field \"{}\" not given.".format(name))

        if name in _fields:
            field = _fields[name]
            type_ = _schema.get('type', "")

            try:
                jsonschema.validate([{"type": type_, "value": field}], TYPE_SCHEMA)
            except jsonschema.exceptions.ValidationError as ex:
                raise ValidationError(ex.message)

            if type_ == 'basic:file:':
                validate_file(field, _schema.get('validate_regex'))

            elif type_ == 'list:basic:file:':
                for obj in field:
                    validate_file(obj, _schema.get('validate_regex'))

            elif type_ == 'basic:dir:':
                validate_dir(field)

            elif type_ == 'list:basic:dir:':
                for obj in field:
                    validate_dir(obj)

            elif type_ == 'basic:json:' and not Storage.objects.filter(pk=field).exists():
                raise ValidationError(
                    "Referenced `Storage` object does not exist (id:{})".format(field))

            elif type_.startswith('data:'):
                validate_data(field, type_)

            elif type_.startswith('list:data:'):
                for data_id in field:
                    validate_data(data_id, type_[5:])  # remove `list:` from type

    for field_schema, fields in iterate_fields(instance, schema):
        pass  # check that schema definitions exist for all fields


def _hydrate_values(output, output_schema, data):
    """Hydrate basic:file and basic:json values.

    Find fields with basic:file type and assign a full path to the file.
    Find fields with basic:json type and assign a JSON object from storage.

    """
    def hydrate_path(file_name):
        id_ = "{}/".format(data.id)  # needs trailing slash
        if id_ in file_name:
            file_name = file_name[file_name.find(id_) + len(id_):]  # remove id from filename

        return os.path.join(settings.FLOW_EXECUTOR['DATA_DIR'], id_, file_name)

    def hydrate_storage(storage_id):
        return LazyStorageJSON(pk=storage_id)

    for field_schema, fields in iterate_fields(output, output_schema):
        name = field_schema['name']
        value = fields[name]
        if 'type' in field_schema:
            if field_schema['type'].startswith('basic:file:'):
                value['file'] = hydrate_path(value['file'])

            elif field_schema['type'].startswith('list:basic:file:'):
                for obj in value:
                    obj['file'] = hydrate_path(obj['file'])

            if field_schema['type'].startswith('basic:dir:'):
                value['dir'] = hydrate_path(value['dir'])

            elif field_schema['type'].startswith('list:basic:dir:'):
                for obj in value:
                    obj['dir'] = hydrate_path(obj['dir'])

            elif field_schema['type'].startswith('basic:json:'):
                fields[name] = hydrate_storage(value)

            elif field_schema['type'].startswith('list:basic:json:'):
                fields[name] = [hydrate_storage(storage_id) for storage_id in value]


def hydrate_size(data):
    """"Add file and dir sizes.

    Add sizes to ``basic:file:``, ``list:basic:file``, ``basic:dir:``
    and ``list:basic:dir:`` fields.
    """

    def add_file_size(obj):
        if data.status in [Data.STATUS_DONE, Data.STATUS_ERROR] and 'size' in obj:
            return

        path = os.path.join(settings.FLOW_EXECUTOR['DATA_DIR'], str(data.pk), obj['file'])
        if not os.path.isfile(path):
            raise ValidationError("Referenced file does not exist ({})".format(path))

        obj['size'] = os.path.getsize(path)

    def get_dir_size(path):
        total_size = 0
        for dirpath, dirnames, filenames in os.walk(path):
            for f in filenames:
                fp = os.path.join(dirpath, f)
                total_size += os.path.getsize(fp)
        return total_size

    def add_dir_size(obj):
        if data.status in [Data.STATUS_DONE, Data.STATUS_ERROR] and 'size' in obj:
            return

        path = os.path.join(settings.FLOW_EXECUTOR['DATA_DIR'], str(data.pk), obj['dir'])
        if not os.path.isdir(path):
            raise ValidationError("Referenced dir does not exist ({})".format(path))

        obj['size'] = get_dir_size(path)

    for field_schema, fields in iterate_fields(data.output, data.process.output_schema):
        name = field_schema['name']
        value = fields[name]
        if 'type' in field_schema:
            if field_schema['type'].startswith('basic:file:'):
                add_file_size(value)
            elif field_schema['type'].startswith('list:basic:file:'):
                for obj in value:
                    add_file_size(obj)
            elif field_schema['type'].startswith('basic:dir:'):
                add_dir_size(value)
            elif field_schema['type'].startswith('list:basic:dir:'):
                for obj in value:
                    add_dir_size(obj)


def hydrate_input_uploads(input_, input_schema, hydrate_values=True):
    """Hydrate input basic:upload types with upload location

    Find basic:upload fields in input.
    Add the upload location for relative paths.

    """
    files = []
    for field_schema, fields in iterate_fields(input_, input_schema):
        name = field_schema['name']
        value = fields[name]
        if 'type' in field_schema:
            if field_schema['type'] == 'basic:file:':
                files.append(value)

            elif field_schema['type'] == 'list:basic:file:':
                files.extend(value)

    urlregex = re.compile(r'^(https?|ftp)://[-A-Za-z0-9\+&@#/%?=~_|!:,.;]*[-A-Za-z0-9\+&@#/%=~_|]')
    for value in files:
        if 'file_temp' in value:
            if isinstance(value['file_temp'], six.string_types):
                # If file_temp not url, nor absolute path: hydrate path
                if not os.path.isabs(value['file_temp']) and not urlregex.search(value['file_temp']):
                    value['file_temp'] = os.path.join(settings.FLOW_EXECUTOR['UPLOAD_DIR'], value['file_temp'])
            else:
                # Something very strange happened
                value['file_temp'] = 'Invalid value for file_temp in DB'


def hydrate_input_references(input_, input_schema, hydrate_values=True):
    """Hydrate ``input_`` with linked data.

    Find fields with complex data:<...> types in ``input_``.
    Assign an output of corresponding data object to those fields.

    """
    for field_schema, fields in iterate_fields(input_, input_schema):
        name = field_schema['name']
        value = fields[name]
        if 'type' in field_schema:
            if field_schema['type'].startswith('data:'):
                # if re.match('^[0-9a-fA-F]{24}$', str(value)) is None:
                #     print "ERROR: data:<...> value in field \"{}\", type \"{}\" not ObjectId but {}.".format(
                #         name, field_schema['type'], value)
                if value is None:
                    continue

                data = Data.objects.get(id=value)
                output = data.output.copy()
                # static = Data.static.to_python(data.static)
                if hydrate_values:
                    _hydrate_values(output, data.process.output_schema, data)
                    # _hydrate_values(static, data.static_schema, data)
                output["__id"] = data.id
                output["__type"] = data.process.type
                fields[name] = output

            elif field_schema['type'].startswith('list:data:'):
                outputs = []
                for val in value:
                    # if re.match('^[0-9a-fA-F]{24}$', str(val)) is None:
                    #     print "ERROR: data:<...> value in {}, type \"{}\" not ObjectId but {}.".format(
                    #         name, field_schema['type'], val)
                    if val is None:
                        continue

                    data = Data.objects.get(id=val)
                    output = data.output.copy()
                    # static = Data.static.to_python(data.static)
                    if hydrate_values:
                        _hydrate_values(output, data.process.output_schema, data)
                        # _hydrate_values(static, data.static_schema, data)

                    output["__id"] = data.id
                    output["__type"] = data.process.type
                    outputs.append(output)

                fields[name] = outputs


def dict_dot(d, k, val=None, default=None):
    """Get or set value using a dot-notation key in a multilevel dict."""
    if val is None and k == '':
        return d

    def set_default(dict_or_model, key, default_value):
        if isinstance(dict_or_model, models.Model):
            if not hasattr(dict_or_model, key):
                setattr(dict_or_model, key, default_value)

            return getattr(dict_or_model, key)
        else:
            return dict_or_model.setdefault(key, default_value)

    def get_item(dict_or_model, key):
        if isinstance(dict_or_model, models.Model):
            return getattr(dict_or_model, key)
        else:
            return dict_or_model[key]

    def set_item(dict_or_model, key, value):
        if isinstance(dict_or_model, models.Model):
            setattr(dict_or_model, key, value)
        else:
            dict_or_model[key] = value

    if val is None and callable(default):
        # Get value, default for missing
        return functools.reduce(lambda a, b: set_default(a, b, default()), k.split('.'), d)

    elif val is None:
        # Get value, error on missing
        return functools.reduce(lambda a, b: get_item(a, b), k.split('.'), d)

    else:
        # Set value
        try:
            k, k_last = k.rsplit('.', 1)
            set_item(dict_dot(d, k, default=dict), k_last, val)
        except ValueError:
            set_item(d, k, val)
        return val



"""
Tests for autoprocess Sphinx extension.

Current implementation of tests only checks that:

    * build process is executed with zero exit status
    * desired html file is produced
    * no errors or warnings were raised during build process

"""
# pylint: disable=missing-docstring, protected-access

import os
import unittest
import shutil

from sphinx import build_main


class TestAutoporocess(unittest.TestCase):

    def setUp(self):
        self.rst_source_files_dir = os.path.dirname(__file__) + "/files/"
        self.build_output_dir = os.path.dirname(__file__) + "/build"
        self.stderr_file = os.path.abspath(os.path.dirname(__file__) + "/files/errors.txt")

    def tearDown(self):
        # Remove the build folder and its contents
        shutil.rmtree(self.build_output_dir)

    def test_build_ok(self):
        args = ['build_sphinx',
                '-E',  # Dont use a saved environment but rebuild it completely.
                '-q',  # Do not output anything on standard output, only write warnings and errors to standard error.
                '-w{}'.format(self.stderr_file),  # Write warnings (and errors) to the given file
                self.rst_source_files_dir,  # souce dir of rst files
                self.build_output_dir  # output dir for html files
                ]

        exit_status = build_main(args)

        self.assertEqual(exit_status, 0)
        # Confirm that html file was generated:
        self.assertTrue(os.path.isfile(os.path.join(self.build_output_dir, 'contents.html')))
        #  Confirm there is no content (no errors/warnings) in self.stderr_file:
        self.assertEqual(os.stat(self.stderr_file).st_size, 0)

if __name__ == '__main__':
    unittest.main()
